@@ -17,6 +17,7 @@ package org.mortbay.jetty.client;
 import java.io.IOException;
 import java.io.InputStream;
 import java.io.InterruptedIOException;
+import java.util.concurrent.atomic.AtomicBoolean;
 
 import org.mortbay.io.Buffer;
 import org.mortbay.io.Buffers;
@@ -49,14 +50,14 @@ public class HttpConnection implements Connection
     boolean _http11 = true;
     Buffer _connectionHeader;
     Buffer _requestContentChunk;
-    long _last;
     boolean _requestComplete;
     public String _message;
     public boolean _reserved;
-
-    /* The current exchange waiting for a response */
+    // The current exchange waiting for a response
     volatile HttpExchange _exchange;
     HttpExchange _pipeline;
+    private final Timeout.Task _timeout = new TimeoutTask();
+    private AtomicBoolean _idle = new AtomicBoolean(false);
 
     public void dump() throws IOException
     {
@@ -68,44 +69,6 @@ public class HttpConnection implements Connection
             ((SslHttpChannelEndPoint)_endp).dump();
     }
 
-    Timeout.Task _timeout = new Timeout.Task()
-    {
-        public void expired()
-        {
-            HttpExchange ex=null;
-            try
-            {
-                synchronized (HttpConnection.this)
-                {
-                    ex = _exchange;
-                    _exchange = null;
-                    if (ex != null)
-                        _destination.returnConnection(HttpConnection.this,true);
-                }
-            }
-            catch (Exception e)
-            {
-                Log.debug(e);
-            }
-            finally
-            {
-                try
-                {
-                    close();
-                }
-                catch (IOException e)
-                {
-                    Log.ignore(e);
-                }
-
-                if (ex!=null && ex.getStatus() < HttpExchange.STATUS_COMPLETED)
-                {
-                    ex.setStatus(HttpExchange.STATUS_EXPIRED);
-                }
-            }
-        }
-    };
-
     /* ------------------------------------------------------------ */
     HttpConnection(Buffers buffers, EndPoint endp, int hbs, int cbs)
     {
@@ -155,19 +118,19 @@ public class HttpConnection implements Connection
             if (!_endp.isOpen())
                 return false;
 
-            ex.setStatus(HttpExchange.STATUS_WAITING_FOR_COMMIT);
             _exchange = ex;
+            _exchange.setStatus(HttpExchange.STATUS_WAITING_FOR_COMMIT);
 
             if (_endp.isBlocking())
+            {
                 this.notify();
+            }
             else
             {
                 SelectChannelEndPoint scep = (SelectChannelEndPoint)_endp;
                 scep.scheduleWrite();
             }
-
-            if (!_endp.isBlocking())
-                _destination.getHttpClient().schedule(_timeout);
+            _destination.getHttpClient().schedule(_timeout);
 
             return true;
         }
@@ -204,10 +167,9 @@ public class HttpConnection implements Connection
                         _parser.skipCRLF();
                         if (_parser.isMoreInBuffer())
                         {
-                            Log.warn("unexpected data");
+                            Log.warn("Unexpected data received but no request sent");
                             close();
                         }
-
                         return;
                     }
                 }
@@ -262,6 +224,11 @@ public class HttpConnection implements Connection
                     }
                 }
 
+                if (_generator.isComplete() && !_requestComplete)
+                {
+                    _requestComplete = true;
+                    _exchange.getEventListener().onRequestComplete();
+                }
 
                 // If we are not ended then parse available
                 if (!_parser.isComplete() && _generator.isCommitted())
@@ -285,27 +252,30 @@ public class HttpConnection implements Connection
             }
             catch (Throwable e)
             {
+                Log.debug("Failure on " + _exchange, e);
+
                 if (e instanceof ThreadDeath)
                     throw (ThreadDeath)e;
-                
+
                 synchronized (this)
                 {
                     if (_exchange != null)
                     {
-                        _exchange.getEventListener().onException(e);
                         _exchange.setStatus(HttpExchange.STATUS_EXCEPTED);
+                        _exchange.getEventListener().onException(e);
                     }
                 }
+
                 failed = true;
                 if (e instanceof IOException)
                     throw (IOException)e;
- 
+
                 if (e instanceof Error)
                     throw (Error)e;
-                
+
                 if (e instanceof RuntimeException)
                     throw (RuntimeException)e;
-                
+
                throw new RuntimeException(e);
             }
             finally
@@ -341,8 +311,8 @@ public class HttpConnection implements Connection
                             close = shouldClose();
 
                         reset(true);
+
                         no_progress = 0;
-                        flushed = -1;
                         if (_exchange != null)
                         {
                             _exchange = null;
@@ -351,8 +321,6 @@ public class HttpConnection implements Connection
                             {
                                 if (!isReserved())
                                     _destination.returnConnection(this,close);
-                                if (close)
-                                    return;
                             }
                             else
                             {
@@ -360,15 +328,17 @@ public class HttpConnection implements Connection
                                 {
                                     if (!isReserved())
                                         _destination.returnConnection(this,close);
-                                    _destination.send(_pipeline);
+
+                                    HttpExchange exchange = _pipeline;
                                     _pipeline = null;
-                                    return;
+                                    _destination.send(exchange);
+                                }
+                                else
+                                {
+                                    HttpExchange exchange = _pipeline;
+                                    _pipeline = null;
+                                    send(exchange);
                                 }
-
-                                HttpExchange ex = _pipeline;
-                                _pipeline = null;
-
-                                send(ex);
                             }
                         }
                     }
@@ -445,11 +415,6 @@ public class HttpConnection implements Connection
             else
             {
                 _exchange._requestFields.remove(HttpHeaders.CONTENT_LENGTH); // TODO
-                // :
-                // should
-                // not
-                // be
-                // needed
                 _generator.completeHeader(_exchange._requestFields,true);
             }
 
@@ -556,39 +521,85 @@ public class HttpConnection implements Connection
         return toString() + " ex=" + _exchange + " " + _timeout.getAge();
     }
 
-    /* ------------------------------------------------------------ */
-    /**
-     * @return the last
-     */
-    public long getLast()
-    {
-        return _last;
-    }
-
-    /* ------------------------------------------------------------ */
-    /**
-     * @param last
-     *            the last to set
-     */
-    public void setLast(long last)
-    {
-        _last = last;
-    }
-
     /* ------------------------------------------------------------ */
     public void close() throws IOException
     {
-        try
-        {
-            _endp.close();
-        }
-        finally
-        {
+        _endp.close();
+    }
 
-            if (_exchange != null)
+
+    /* ------------------------------------------------------------ */
+    public void setIdleTimeout(long expire)
+    {
+        synchronized (this)
+        {
+            if (_idle.compareAndSet(false,true))
+                _destination.getHttpClient().scheduleIdle(_timeout);
+            else
+                throw new IllegalStateException();
+        }
+    }
+    
+    /* ------------------------------------------------------------ */
+    public boolean cancelIdleTimeout()
+    {
+        synchronized (this)
+        {
+            if (_idle.compareAndSet(true,false))
             {
-                _exchange.onException(new IOException("Connection closed"));
+                _destination.getHttpClient().cancel(_timeout);
+                return true;
+            }
+        }
+        
+        return false;
+    }
+
+    /* ------------------------------------------------------------ */
+    /* ------------------------------------------------------------ */
+    /* ------------------------------------------------------------ */
+    private class TimeoutTask extends Timeout.Task
+    {
+        public void expired()
+        {
+            HttpExchange ex=null;
+            try
+            {
+                synchronized (HttpConnection.this)
+                {
+                    ex = _exchange;
+                    _exchange = null;
+                    if (ex != null)
+                    {
+                        _destination.returnConnection(HttpConnection.this,true);
+                    }
+                    else if (_idle.compareAndSet(true,false))
+                    {
+                        _destination.returnIdleConnection(HttpConnection.this);
+                    }
+                }
+            }
+            catch (Exception e)
+            {
+                Log.debug(e);
+            }
+            finally
+            {
+                try
+                {
+                    close();
+                }
+                catch (IOException e)
+                {
+                    Log.ignore(e);
+                }
+
+                if (ex!=null && ex.getStatus() < HttpExchange.STATUS_COMPLETED)
+                {
+                    ex.setStatus(HttpExchange.STATUS_EXPIRED);
+                }
             }
         }
     }
+
 }
