@@ -20,6 +20,17 @@
  */
 package org.mortbay.jetty;
 
+import java.io.IOException;
+import java.io.PrintWriter;
+
+
+import javax.servlet.ServletException;
+import javax.servlet.http.HttpServletResponse;
+import javax.servlet.http.HttpServletRequest;
+
+import org.mortbay.log.Log;
+
+
 import junit.framework.TestCase;
 
 /**
@@ -340,6 +351,62 @@ public class HttpConnectionTest extends TestCase
         }
     }
     
+    
+    public void testOversizedResponse ()
+    throws Exception
+    {  
+        String str = "thisisastringthatshouldreachover1kbytes";
+        for (int i=0;i<400;i++)
+            str+="xxxxxxxxxxxx";
+        final String longstr = str;
+        String response = null;
+        server.stop();
+        server.setHandler(new DumpHandler()
+        {
+            public void handle(String target, HttpServletRequest request, HttpServletResponse response, int dispatch) throws IOException, ServletException
+            {
+                try
+                {
+                    Request base_request = (request instanceof Request) ? (Request)request:HttpConnection.getCurrentConnection().getRequest();
+                    base_request.setHandled(true);
+                    response.setHeader(HttpHeaders.CONTENT_TYPE,MimeTypes.TEXT_HTML);
+                    response.setHeader("LongStr", longstr);
+                    PrintWriter writer = response.getWriter();
+                    writer.write("<html><h1>FOO</h1></html>");  
+                    writer.flush();
+                    writer.close();
+                    throw new RuntimeException("SHOULD NOT GET HERE");
+                }
+                catch(ArrayIndexOutOfBoundsException e)
+                {
+                    Log.debug(e);
+                    Log.info("correctly ignored "+e);
+                }
+            }
+        });
+        server.start();
+        
+        connector.reopen();
+        try 
+        {
+            int offset = 0;
+          
+            response = connector.getResponses("GET / HTTP/1.1\n"+
+                "Host: localhost\n" +
+                "\015\012"
+             );
+          
+            offset = checkContains(response, offset, "HTTP/1.1 500");
+        } 
+        catch(Exception e)
+        {
+            e.printStackTrace();
+            if(response != null)
+                System.err.println(response);
+            fail("Exception");      
+        }
+    }
+    
     public void testAsterisk()
     {
         String response = null;
