@@ -497,17 +497,18 @@ public class UrlEncoded extends MultiMap
     public static void decodeTo(InputStream in, MultiMap map, String charset, int maxLength)
     throws IOException
     {
-        if (charset==null || StringUtil.__ISO_8859_1.equals(charset))
-        {
-            decode88591To(in,map,maxLength);
-            return;
-        }
 
-        if (StringUtil.__UTF8.equalsIgnoreCase(charset))
+        if (charset==null || StringUtil.__UTF8.equalsIgnoreCase(charset))
         {
             decodeUtf8To(in,map,maxLength);
             return;
         }
+        
+        if (StringUtil.__ISO_8859_1.equals(charset))
+        {
+            decode88591To(in,map,maxLength);
+            return;
+        }
 
         if (StringUtil.__UTF16.equalsIgnoreCase(charset)) // Should be all 2 byte encodings
         {
